/*
  xrus - keyboard switcher/indicator
  Copyright (c) 1996 Alexander V. Lukyanov
  This is free software with no warranty.
  See COPYING for details.
*/
/*__________________________________________________________________________
**
**    File:    xrus.c
**__________________________________________________________________________
*/

#include <config.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <errno.h>
#include <time.h>
#include <ctype.h>
#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif
#ifdef HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif
#ifdef __GNUC__
# ifndef alloca
#  define alloca __builtin_alloca
# endif
#else
# ifdef HAVE_ALLOCA_H
#  include <alloca.h>
# endif
#endif

#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <Xm/Xm.h>
#include <Xm/PushB.h>
#include <Xm/Form.h>
#include <Xm/MwmUtil.h>

#define  XK_MISCELLANY
#include <X11/keysymdef.h>

#ifndef SA_RESTART
#define SA_RESTART 0
#endif

#include "keycomb.h"
#include "xrus.h"
#include "props.h"
#include "menu.h"

const char  AppClass[]="Xrus";

extern   String   DefaultResources;

XtActionsRec   Actions[]=
{
   {"XrusMenuPopup",XrusMenuPopup}
};

#define  SwappableKey(k)   ((k)[0]>0&&(k)[1]>0&&(k)[2]>0&&(k)[3]>0&&\
                            (k)[0]<256&&(k)[1]<256&&(k)[2]<256&&(k)[3]<256)

struct wincache_cell
{
   Window   win;
   time_t   time;
}
   window_cache[256];

XrusRec  AppData;

char  DefaultSwitchKeys[]="Shift_L+Shift_R";
char  DefaultToLatKeys[]="";
char  DefaultToRusKeys[]="";
char  DefaultLocker[]="exec xlock -remote >/dev/null 2>&1";

char  LockerData[256];
char  SwitchKeysData[256];
char  ToLatKeysData[256];
char  ToRusKeysData[256];

XtResource        resources[]=
{
/* {  name        class       type        size
         offset
         default_type default_addr              },*/
/* {  ----------- ----------- ----------- ------------
         -----------------------------------
         ----------- -------------------------  },*/
   {  "autolock", "Autolock", XtRBoolean, sizeof(Boolean),
         XtOffsetOf(XrusRec,autolock),
         XtRImmediate, (XtPointer)False         },
   {  "locker",   "Locker",   XtRString,  sizeof(String),
         XtOffsetOf(XrusRec,locker),
         XtRString,  DefaultLocker              },
   {  "timeout",  "Timeout",  XtRInt,     sizeof(int),
         XtOffsetOf(XrusRec,timeout),
         XtRImmediate,  (XtPointer)30           },
   {  "useBell", "UseBell",   XtRBoolean, sizeof(int),
         XtOffsetOf(XrusRec,useBell),
         XtRImmediate, (XtPointer)True          },
   {  "switchKeys","SwitchKeys",XtRString,sizeof(String),
         XtOffsetOf(XrusRec,switchKeys),
         XtRString,  DefaultSwitchKeys          },
   {  "toRusKeys","ToRusKeys",XtRString,sizeof(String),
         XtOffsetOf(XrusRec,toRusKeys),
         XtRString,  DefaultToRusKeys           },
   {  "toLatKeys","ToLatKeys",XtRString,sizeof(String),
         XtOffsetOf(XrusRec,toLatKeys),
         XtRString,  DefaultToLatKeys           },
   {  "xmodmap",  "Xmodmap",  XtRString,  sizeof(String),
         XtOffsetOf(XrusRec,xmodmap),
         XtRString,  ""                         },
   {  "led",      "Led",      XtRInt,     sizeof(int),
         XtOffsetOf(XrusRec,led),
         XtRImmediate,  (XtPointer)0            },
   {  "altMaps",  "AltMaps",  XtRString,  sizeof(String),
         XtOffsetOf(XrusRec,altMaps),
         XtRString,  ""                         }
};

XrmOptionDescRec  options[]=
{
/* {  option,       specifier,     argKind,          value  },*/
/* {  ------------- -------------- ----------------- ------ },*/
   {  "-autolock",  "*autolock",   XrmoptionNoArg,   "true" },
   {  "-locker",    "*locker",     XrmoptionSepArg,  NULL   },
   {  "-timeout",   "*timeout",    XrmoptionSepArg,  NULL   },
   {  "-bell",      "*useBell",    XrmoptionNoArg,   "true" },
   {  "+bell",      "*useBell",    XrmoptionNoArg,   "false"},
   {  "-led",       "*led",        XrmoptionSepArg,  "0"   },
};

Display        *disp;
XtAppContext   app_context;

#define	StartArgs()		count=0
#define	AddArg(name,val)	(XtSetArg(args[count],(name),(val)),++count)

char  *program;

struct KeyCombination   SwitchKeys,ToLatKeys,ToRusKeys;

int   FirstKeycode;
int   LastKeycode;

enum
{
   MODE=1,
   RUS=1,
   LAT=0,

   CODING=6,
   KOI=0,
   ALT=2
};
int   Mode=LAT|KOI;

Widget   top_level,form_w;
Widget   switch_button[2];

Atom  type_ret;
int   format_ret;
unsigned long nitems_ret,bytes_after_ret;
unsigned char *prop;

int   xrus_check=1;

void  AddChildren(Window);
void  SetAlarm();

int   (*old_error_handler)(Display*,XErrorEvent*);

pid_t LockerRunning=0;

int   MappingNotifyIgnoreCount=0;
XtIntervalId   MappingNotifyTimeout=-1;

/*________________________________________________________________________
*/

void  SwitchKeyboard(int from,int to)
{
   int      ks_per_kc;
   KeySym   *keymap;
   KeySym   tmp;
   int      i;
   XKeyboardControl  kbd_control;

   if(from==to)
      return;

   keymap=XGetKeyboardMapping(disp,FirstKeycode,LastKeycode-FirstKeycode+1,&ks_per_kc);

   if(ks_per_kc<4)
      return;

   for(i=(LastKeycode-FirstKeycode)*ks_per_kc; i>=0; i-=ks_per_kc)
   {
      if((from&MODE)!=(to&MODE))
      {
         if(SwappableKey(keymap+i))
         {
            tmp=keymap[i];
            keymap[i]=keymap[i+2];
            keymap[i+2]=tmp;
            tmp=keymap[i+1];
            keymap[i+1]=keymap[i+3];
            keymap[i+3]=tmp;
         }
      }
   }

   XChangeKeyboardMapping(disp,FirstKeycode,ks_per_kc,keymap,LastKeycode-FirstKeycode+1);
   MappingNotifyIgnoreCount++;
   if(AppData.led>0 && AppData.led<=32)
   {
      kbd_control.led_mode=((to&MODE)==RUS?LedModeOn:LedModeOff);
      kbd_control.led=AppData.led;
      XChangeKeyboardControl(disp,KBLed|KBLedMode,&kbd_control);
   }

   XFree(keymap);

#ifdef sgi
   {
      /* This might seem strange, but SGI wants this */
      XModifierKeymap *map=XGetModifierMapping(disp);
      XSetModifierMapping(disp,map);
      XFreeModifiermap(map);
   }
#endif

   Mode=to;
}

/* shows proper indicator button */
void  ShowSwitchButton(void)
{
   int   to_manage=-1;

   if((Mode&MODE)==RUS && !XtIsManaged(switch_button[1]))
      to_manage=1;
   else if((Mode&MODE)==LAT && !XtIsManaged(switch_button[0]))
      to_manage=0;
   if(to_manage!=-1)
   {
      Arg   args[1];

      XtSetArg(args[0],XtNallowShellResize,False);
      XtSetValues(top_level,args,1);

      XtUnmanageChild(switch_button[1-to_manage]);

      XtSetArg(args[0],XtNallowShellResize,True);
      XtSetValues(top_level,args,1);

      XtManageChild(switch_button[to_manage]);
   }
}

void  PushSwitchButtonCB(Widget w,XtPointer closure,XtPointer call_data)
{
   SwitchKeyboard(Mode,Mode^MODE);

   if(AppData.useBell)
   {
      XSync(disp,False);
      XBell(disp,30);
   }
   XRaiseWindow(disp,XtWindow(top_level));
   ShowSwitchButton();
}

/* a window is in cache => we have requested all needed input from the window */
/* returns a boolean indicating the window was in cache
   and adds it to cache if it was not */
int   AddToWindowCache(Window w)
{
   time_t   t=time(NULL);
   time_t   best_time=t;
   int      best_cell=-1;
   int      i;

   for(i=0; i<XtNumber(window_cache); i++)
   {
      if(window_cache[i].win==w && window_cache[i].time!=0)
      {
         window_cache[i].time=t;
         return(0);
      }
      if(best_cell==-1 || best_time>window_cache[i].time || window_cache[i].win==0)
      {
         best_cell=i;
         best_time=window_cache[i].time;
      }
   }
   window_cache[best_cell].time=t;
   window_cache[best_cell].win=w;
   return(1);
}

/* deletes a window from the cache */
void  DelWindow(Window w)
{
   int   i;

   for(i=0; i<XtNumber(window_cache); i++)
   {
      if(window_cache[i].win==w)
      {
         window_cache[i].win=0;
         window_cache[i].time=0;
      }
   }
}

int   window_is_top_level(Window w)
{
   XGetWindowProperty(disp,w,XA_WM_CLASS,0L,0L,0,XA_STRING,
         &type_ret,&format_ret,&nitems_ret,&bytes_after_ret,&prop);
   return(type_ret!=None);
}

int   is_root_child(Window root,Window w)
{
   Window   root1,parent,*children;
   unsigned children_num,i;

   if(XQueryTree(disp,root,&root1,&parent,&children,&children_num))
   {
      for(i=0; i<children_num; i++)
         if(children[i]==w)
            break;
      XFree(children);
      if(i<children_num)
         return True;
   }
   return False;
}

void  AddWindow(Window w)
{
   XWindowAttributes wa;

   if(AddToWindowCache(w)) /* if the window is not in cache */
   {
      if(!XGetWindowAttributes(disp,w,&wa))
      {
         DelWindow(w);
         return;
      }
      /* we want to be notified of subwindows creation/destruction etc */
      wa.your_event_mask|=SubstructureNotifyMask;

      /* we request events only if
         1. the window already requested this kind of events
         2. the window is blocking the events from propagation
         3. this window is wm's one or top level one
            (for the case the application itself doesn't want these events)
      */
      wa.your_event_mask|=(wa.all_event_masks|wa.do_not_propagate_mask)
                          &(KeyPressMask|KeyReleaseMask|PointerMotionMask);

      /* we always want both Press and Release events */
      if(wa.your_event_mask&(KeyPressMask|KeyReleaseMask))
         wa.your_event_mask|=(KeyPressMask|KeyReleaseMask);

      /* third condition */
      if((wa.your_event_mask&(KeyPressMask|KeyReleaseMask|PointerMotionMask))
         !=(KeyPressMask|KeyReleaseMask|PointerMotionMask)
      && (w==wa.root || window_is_top_level(w) || is_root_child(wa.root,w)))
         wa.your_event_mask|=(KeyPressMask|KeyReleaseMask|PointerMotionMask);
      else
         DelWindow(w);  /* we'll check input mask later */

      XSelectInput(disp,w,wa.your_event_mask);

      AddChildren(w);
   }
}

void  AddChildren(Window w)
{
   Window   root,parent,*children;
   unsigned children_num,i;
   Status   st;

   if(xrus_check)
   {
      st=XGetWindowProperty(disp,w,XA_WM_CLASS,0L,256L,0,XA_STRING,
            &type_ret,&format_ret,&nitems_ret,&bytes_after_ret,&prop);
      if(st && type_ret!=None)
      {
         if(!strcmp(prop+strlen(prop)+1,AppClass))
         {
            fprintf(stderr,"%s: xrus already runs on the display\n",program);
            exit(1);
         }
         XtFree(prop);
      }
   }

   if(XQueryTree(disp,w,&root,&parent,&children,&children_num))
   {
      for(i=0; i<children_num; i++)
         AddWindow(children[i]);
      XFree(children);
   }
}

void  SwitchKeysFire()
{
   PushSwitchButtonCB(switch_button[(Mode&MODE)==RUS],(XtPointer)(switch_button[(Mode&MODE)!=RUS]),NULL);
}
void  ToLatKeysFire()
{
   if((Mode&MODE)==RUS)
      SwitchKeysFire();
}
void  ToRusKeysFire()
{
   if((Mode&MODE)==LAT)
      SwitchKeysFire();
}

void  CheckKeymap()
{
   int      ks_per_kc;
   KeySym   *keymap;
   int      ok_rus=0;
   int      ok_lat=0;
   int      i;

   keymap=XGetKeyboardMapping(disp,FirstKeycode,LastKeycode-FirstKeycode+1,&ks_per_kc);

   if(ks_per_kc>=4)
   {
      for(i=(LastKeycode-FirstKeycode)*ks_per_kc; i>=0; i-=ks_per_kc)
      {
         if(keymap[i] && keymap[i+1] && keymap[i+2] && keymap[i+3])
         {
            ok_lat|=(keymap[i+2]>127 && keymap[i+2]<256 && keymap[i+3]>127 && keymap[i+3]<256);
            ok_rus|=(keymap[i+0]>127 && keymap[i+0]<256 && keymap[i+1]>127 && keymap[i+1]<256);
         }
      }
   }

   XFree(keymap);

   if(ok_rus==ok_lat)
   {
      fprintf(stderr,"%s: key mapping is not in proper coding, exiting\n",program);
      exit(1);
   }
   Mode=(Mode&~MODE)|(ok_rus?RUS:LAT);
   if(switch_button[0] && switch_button[1])
      ShowSwitchButton();
}

void  MappingNotifyTimeoutHandler(XtPointer closure,XtIntervalId *id)
{
   (void)closure; (void)id;
   MappingNotifyTimeout=-1;
   CheckKeymap();
}

void  MainLoop(void)
{
   XEvent   ev;
   int      i;

   for(;;)
   {
      if(app_context==NULL)
         return;
      XtAppNextEvent(app_context,&ev);
      XtDispatchEvent(&ev);
      switch(ev.type)
      {
      case(CreateNotify):
         AddWindow(ev.xcreatewindow.window);
         break;
      case(MapNotify):
         /* just to be sure we know the latest input mask */
         AddWindow(ev.xmap.window);
         break;
      case(DestroyNotify):
         DelWindow(ev.xdestroywindow.window);
         break;
      case(KeyPress):
         SetAlarm();
         KeyCombinationProcessKeyPress(&SwitchKeys,&ev.xkey);
         KeyCombinationProcessKeyPress(&ToRusKeys,&ev.xkey);
         KeyCombinationProcessKeyPress(&ToLatKeys,&ev.xkey);
         break;
      case(KeyRelease):
         SetAlarm();
         i=KeyCombinationProcessKeyRelease(&SwitchKeys,&ev.xkey,1);
         i|=KeyCombinationProcessKeyRelease(&ToLatKeys,&ev.xkey,!i);
         KeyCombinationProcessKeyRelease(&ToRusKeys,&ev.xkey,!i);
         break;
      case(MotionNotify):
      case(ButtonPress):
      case(ButtonRelease):
         SetAlarm();
         break;
      case(MappingNotify):
         if(ev.xmapping.request==MappingKeyboard)
         {
            /* workaround with timeout for xmodmap bug (too many events) */
            if(MappingNotifyIgnoreCount==0)
            {
               if(MappingNotifyTimeout!=-1)
                  XtRemoveTimeOut(MappingNotifyTimeout);
               MappingNotifyTimeout=XtAppAddTimeOut(app_context,200,
                                       MappingNotifyTimeoutHandler,NULL);
            }
            else
               MappingNotifyIgnoreCount--;
         }
         break;
      }
   }
}

void  cleanup(void)
{
   if(app_context)
      SwitchKeyboard(Mode,LAT|KOI);
}

void  sig_term(int sig)
{
   fprintf(stderr,"%s: caught signal %d, exiting\n",program,sig);
   exit(1);
}

void  sig_chld(int sig)
{
   int   status;
   if(wait(&status)==LockerRunning)
   {
      LockerRunning=0;
      SetAlarm();
   }
}

void  LockScreen()
{
   pid_t pid;

   ToLatKeysFire();

   fflush(stderr);
   switch(pid=fork())
   {
   case(0):
      execl("/bin/sh","sh","-c",AppData.locker,NULL);
      fprintf(stderr,"%s: cannot execute /bin/sh - %s\n",program,strerror(errno));
      fflush(stderr);
      _exit(1);
   case(-1):
      fprintf(stderr,"%s: cannot do fork() - %s\n",program,strerror(errno));
      fflush(stderr);
      break;
   default:
      LockerRunning=pid;
   }
}

void  AlarmHandler(int sig)
{
   (void)sig;
   if(AppData.autolock)
      LockScreen();
}

void  SetAlarm()
{
   if(AppData.autolock && AppData.timeout>0 && !LockerRunning)
      alarm(AppData.timeout*60);
   else
      alarm(0);
}

int   X_error_handler(Display *d,XErrorEvent *ev)
{
   /* A window can be deleted between CreateNotify and our request */

   if(ev->error_code==BadWindow
   || ev->error_code==BadDrawable
   || ev->error_code==BadAccess)
      return(-1);
   return(old_error_handler(d,ev));
}

int   main(int argc,char **argv)
{
   Arg   args[32];
   int   count=0;
   Dimension   width,height;
   Atom  ol_decor_del_atom,ol_decor_del[2];
   struct sigaction  act;
   XrmDatabase default_rdb,rdb;
   char  *HOME;
   char  *buf;
   char  err[1024];
   int   scr;

   printf("Xrus | Version " VERSION " | Copyright (c) 1995-96 Alexander V. Lukyanov\n");
   fflush(stdout);

   program=argv[0];

   act.sa_handler=sig_term;
   sigemptyset(&act.sa_mask);
   act.sa_flags=0;
   sigaction(SIGTERM,&act,NULL);
   sigaction(SIGHUP,&act,NULL);
   sigaction(SIGINT,&act,NULL);
   sigaction(SIGQUIT,&act,NULL);

   act.sa_handler=AlarmHandler;
   act.sa_flags=SA_RESTART;
   sigaction(SIGALRM,&act,NULL);
   act.sa_handler=sig_chld;
   act.sa_flags=SA_RESTART|SA_NOCLDSTOP;
   sigaction(SIGCHLD,&act,NULL);

   XtToolkitInitialize();
   app_context=XtCreateApplicationContext();
   disp=XtOpenDisplay(app_context,NULL,"xrus",AppClass,options,XtNumber(options),&argc,argv);
   if(disp==NULL)
   {
      fprintf(stderr,"%s: cannot open display\n",program);
      exit(1);
   }
   default_rdb=XrmGetStringDatabase(DefaultResources);
   rdb=XtDatabase(disp);
   XrmCombineDatabase(default_rdb,&rdb,False);
   HOME=getenv("HOME");
   if(HOME)
   {
      buf=alloca(strlen(HOME)+10);
      sprintf(buf,"%s/.xrus",HOME);
      XrmCombineFileDatabase(buf,&rdb,True);
   }

   old_error_handler=XSetErrorHandler(X_error_handler);

   for(scr=0; scr<ScreenCount(disp); scr++)
   {
      AddWindow(RootWindowOfScreen(ScreenOfDisplay(disp,scr)));
   }
   xrus_check=0;

   StartArgs();
   AddArg(XmNmwmDecorations,MWM_DECOR_BORDER);
   AddArg(XmNinput,False);
   AddArg(XmNallowShellResize,True);
   AddArg(XmNmappedWhenManaged,False);
   top_level=XtAppCreateShell("xrus",AppClass,applicationShellWidgetClass,disp,args,count);

   XtGetApplicationResources(top_level,&AppData,resources,XtNumber(resources),NULL,0);

   strcpy(LockerData,AppData.locker?AppData.locker:DefaultLocker);
   AppData.locker=LockerData;
   strcpy(SwitchKeysData,AppData.switchKeys?AppData.switchKeys:DefaultSwitchKeys);
   AppData.switchKeys=SwitchKeysData;
   strcpy(ToLatKeysData,AppData.toLatKeys?AppData.toLatKeys:DefaultToLatKeys);
   AppData.toLatKeys=ToLatKeysData;
   strcpy(ToRusKeysData,AppData.toRusKeys?AppData.toRusKeys:DefaultToRusKeys);
   AppData.toRusKeys=ToRusKeysData;

   if(argc>1)
   {
      AppData.xmodmap=argv[1];
   }

   if(strlen(AppData.xmodmap)>0)
   {
      char *cmd=alloca(strlen(AppData.xmodmap)+9);
      sprintf(cmd,"xmodmap %s",AppData.xmodmap);
      system(cmd);
   }

   XDisplayKeycodes(disp,&FirstKeycode,&LastKeycode);
   CheckKeymap();

   /* initialize switch keys */
   SwitchKeys.Fire=SwitchKeysFire;
   ParseKeyCombination(AppData.switchKeys,&SwitchKeys,err);
   PrintParseErrors("switchKeys",err);
   ToLatKeys.Fire=ToLatKeysFire;
   ParseKeyCombination(AppData.toLatKeys,&ToLatKeys,err);
   PrintParseErrors("toLatKeys",err);
   ToRusKeys.Fire=ToRusKeysFire;
   ParseKeyCombination(AppData.toRusKeys,&ToRusKeys,err);
   PrintParseErrors("toRusKeys",err);

#if !defined(DEBUG) && !defined(DONT_FORK)
   switch(fork())
   {
   case(0):
      break;
   case(-1):
      perror("fork");
      exit(1);
   default:
      _exit(0);
   }
#endif

   XtAppAddActions(app_context,Actions,XtNumber(Actions));

   StartArgs();
   AddArg(XmNresizable,False);
   form_w=XtCreateManagedWidget("form",xmFormWidgetClass,top_level,args,count);

   switch_button[0]=XtCreateWidget("modeButton0",xmPushButtonWidgetClass,form_w,NULL,0);
   switch_button[1]=XtCreateWidget("modeButton1",xmPushButtonWidgetClass,form_w,NULL,0);

   XtManageChild(switch_button[(Mode&MODE)==RUS]);

   XtAddCallback(switch_button[0],XmNactivateCallback,PushSwitchButtonCB,
                                                (XtPointer)switch_button[1]);
   XtAddCallback(switch_button[1],XmNactivateCallback,PushSwitchButtonCB,
                                                (XtPointer)switch_button[0]);

   ol_decor_del_atom=XInternAtom(disp,"_OL_DECOR_DEL",0);
   ol_decor_del[0]=XInternAtom(disp,"_OL_DECOR_HEADER",0);
   ol_decor_del[1]=XInternAtom(disp,"_OL_DECOR_RESIZE",0);

   XtRealizeWidget(top_level);

   XChangeProperty(disp,XtWindow(top_level),ol_decor_del_atom,XA_ATOM,32,
         PropModeReplace,(void*)ol_decor_del,2);

   atexit(cleanup);

   StartArgs();
   AddArg(XmNwidth,&width);
   AddArg(XmNheight,&height);
   XtGetValues(form_w,args,count);
   StartArgs();
   AddArg(XmNmaxWidth,width);
   AddArg(XmNmaxHeight,height);
   AddArg(XmNminWidth,width);
   AddArg(XmNminHeight,height);
   XtSetValues(top_level,args,count);

   XtMapWidget(top_level);

   XrusMenuCreate();

   SetAlarm();
   MainLoop();

   return 0;
}
